// Copyright 2021 Harran Ali <harran.m@gmail.com>. All rights reserved.
// Copyright (c) 2024 Zeni Kim <zenik@smarteching.com>
// Use of this source code is governed by MIT-style
// license that can be found in the LICENSE file.

package core

import (
	"bytes"
	"fmt"
	"net/http"
)

// Response represents an HTTP response to be sent to the client, including headers, body, status code, and metadata.
type Response struct {
	headers             []header
	body                []byte
	statusCode          int
	contentType         string
	overrideContentType string
	isTerminated        bool
	redirectTo          string
	HttpResponseWriter  http.ResponseWriter
}

// header represents a single HTTP header with a key-value pair.
type header struct {
	key string
	val string
}

// writes the contents of a buffer to the HTTP response with specified file name and type if not terminated.
func (rs *Response) BufferFile(name string, filetype string, b bytes.Buffer) *Response {

	if rs.isTerminated == false {
		rs.HttpResponseWriter.Header().Add("Content-Type", filetype)
		rs.HttpResponseWriter.Header().Add("Content-Disposition", "attachment; filename="+name)
		b.WriteTo(rs.HttpResponseWriter)
	}
	return rs
}

// writes the contents of a buffer to the HTTP response with specified file name and type if not terminated.
func (rs *Response) BufferInline(name string, filetype string, b bytes.Buffer) *Response {

	if rs.isTerminated == false {
		rs.HttpResponseWriter.Header().Add("Content-Type", filetype)
		b.WriteTo(rs.HttpResponseWriter)
	}
	return rs
}

// sets the response's content type to HTML and assigns the provided body as the response body if not terminated.
func (rs *Response) Any(body any) *Response {
	if rs.isTerminated == false {
		rs.contentType = CONTENT_TYPE_HTML
		rs.body = []byte(rs.castBasicVarsToString(body))
	}

	return rs
}

// sets the response's content type to JSON and assigns the provided string as the body if the response is not terminated.
func (rs *Response) Json(body string) *Response {
	if rs.isTerminated == false {
		rs.contentType = CONTENT_TYPE_JSON
		rs.body = []byte(body)
	}
	return rs
}

// sets the response's content type to plain text and assigns the provided string as the body if not terminated.
func (rs *Response) Text(body string) *Response {
	if rs.isTerminated == false {
		rs.contentType = CONTENT_TYPE_TEXT
		rs.body = []byte(body)
	}
	return rs
}

// sets the response's content type to HTML and assigns the provided string as the body if the response is not terminated.
func (rs *Response) HTML(body string) *Response {
	if rs.isTerminated == false {
		rs.contentType = CONTENT_TYPE_HTML
		rs.body = []byte(body)
	}
	return rs
}

// renders the specified template with the provided data and writes the result to the HTTP response if not terminated.
func (rs *Response) Template(name string, data interface{}) *Response {
	var buffer bytes.Buffer
	if rs.isTerminated == false {
		err := tmpl.ExecuteTemplate(&buffer, name, data)
		if err != nil {
			panic(fmt.Sprintf("error executing template: %v", err))
		}
		rs.contentType = CONTENT_TYPE_HTML
		buffer.WriteTo(rs.HttpResponseWriter)
	}
	return rs
}

// sets the response status code if the response is not yet terminated. Returns the updated Response object.
func (rs *Response) SetStatusCode(code int) *Response {
	if rs.isTerminated == false {
		rs.statusCode = code
	}

	return rs
}

// sets a custom content type for the response if it is not terminated. Returns the updated Response object.
func (rs *Response) SetContentType(c string) *Response {
	if rs.isTerminated == false {
		rs.overrideContentType = c
	}

	return rs
}

// adds or updates a header to the response if it is not terminated. Returns the updated Response object.
func (rs *Response) SetHeader(key string, val string) *Response {
	if rs.isTerminated == false {
		h := header{
			key: key,
			val: val,
		}
		rs.headers = append(rs.headers, h)
	}
	return rs
}

// terminates the response processing, preventing any further modifications or actions.
func (rs *Response) ForceSendResponse() {
	rs.isTerminated = true
}

// updates the redirect URL for the response and returns the modified Response. Validates the URL before setting it.
func (rs *Response) Redirect(url string) *Response {
	validator := resolveValidator()
	v := validator.Validate(map[string]interface{}{
		"url": url,
	}, map[string]interface{}{
		"url": "url",
	})
	if v.Failed() {
		if url[0:1] != "/" {
			rs.redirectTo = "/" + url
		} else {
			rs.redirectTo = url
		}
		return rs
	}
	rs.redirectTo = url
	return rs
}

// converts various primitive data types to their string representation or triggers a panic for unsupported types.
func (rs *Response) castBasicVarsToString(data interface{}) string {
	switch dataType := data.(type) {
	case string:
		return fmt.Sprintf("%v", data)
	case []byte:
		d := data.(string)
		return fmt.Sprintf("%v", d)
	case int:
		intVar, _ := data.(int)
		return fmt.Sprintf("%v", intVar)
	case int8:
		int8Var := data.(int8)
		return fmt.Sprintf("%v", int8Var)
	case int16:
		int16Var := data.(int16)
		return fmt.Sprintf("%v", int16Var)
	case int32:
		int32Var := data.(int32)
		return fmt.Sprintf("%v", int32Var)
	case int64:
		int64Var := data.(int64)
		return fmt.Sprintf("%v", int64Var)
	case uint:
		uintVar, _ := data.(uint)
		return fmt.Sprintf("%v", uintVar)
	case uint8:
		uint8Var := data.(uint8)
		return fmt.Sprintf("%v", uint8Var)
	case uint16:
		uint16Var := data.(uint16)
		return fmt.Sprintf("%v", uint16Var)
	case uint32:
		uint32Var := data.(uint32)
		return fmt.Sprintf("%v", uint32Var)
	case uint64:
		uint64Var := data.(uint64)
		return fmt.Sprintf("%v", uint64Var)
	case float32:
		float32Var := data.(float32)
		return fmt.Sprintf("%v", float32Var)
	case float64:
		float64Var := data.(float64)
		return fmt.Sprintf("%v", float64Var)
	case complex64:
		complex64Var := data.(complex64)
		return fmt.Sprintf("%v", complex64Var)
	case complex128:
		complex128Var := data.(complex128)
		return fmt.Sprintf("%v", complex128Var)
	case bool:
		boolVar := data.(bool)
		return fmt.Sprintf("%v", boolVar)
	default:
		panic(fmt.Sprintf("unsupported response data type %v!", dataType))
	}
}

// reinitializes the Response object to its default state, clearing its body, headers, and resetting other properties.
func (rs *Response) reset() {
	rs.body = nil
	rs.statusCode = http.StatusOK
	rs.contentType = CONTENT_TYPE_HTML
	rs.overrideContentType = ""
	rs.isTerminated = false
	rs.redirectTo = ""
}
